
/* shadowvol.c - by Tom McReynolds, SGI */

/* Shadows: Shadow maps */

// Revised version using modern OpenGL and the OpenGL Utilities for Geometry Generation.
// By Ingemar Ragnemalm 2022

// C++ version 2024!

#define MAIN
#include <OpenGL/gl3.h>
#include "MicroGlut.h"
#include "VectorUtils4.h"
#include "glugg.h"
#include "gluggShapes.h"
#include <stdlib.h>
#include "SimpleGUI.h"
// uses framework Cocoa
// uses framework OpenGL

/* Demonstrate shadow volumes */

/* Create a single component texture map */
GLfloat *
make_texture(int maxs, int maxt)
{
  int s, t;
  static GLfloat *texture;

  texture = (GLfloat *) malloc(maxs * maxt * sizeof(GLfloat));
  for (t = 0; t < maxt; t++)
  {
    for (s = 0; s < maxs; s++)
    {
      texture[s + maxs * t] = ((s >> 4) & 0x1) ^ ((t >> 4) & 0x1);
    }
  }
  return texture;
}

/* Create an RGBA texture map */
GLfloat *
make_texture2(int maxs, int maxt)
{
  int s, t;
  static GLfloat *texture;

  texture = (GLfloat *) malloc(maxs * maxt * sizeof(GLfloat) * 4);
  for (t = 0; t < maxt; t++)
  {
    for (s = 0; s < maxs; s++)
    {
      texture[(s + maxs * t)*4 + 0] = (((s >> 4) & 0x1) ^ ((t >> 4) & 0x1));
      texture[(s + maxs * t)*4 + 1] = (((s >> 4) & 0x1) ^ ((t >> 4) & 0x1));
      texture[(s + maxs * t)*4 + 2] = (((s >> 4) & 0x1) ^ ((t >> 4) & 0x1));
      texture[(s + maxs * t)*4 + 3] = 1.0;
    }
  }
  return texture;
}

typedef struct
{
  vec3 *verticies;
  vec3 normal;
  int n;                /* number of verticies */
} ShadObj;

vec3 shadVerts[] =
{
  vec3(30.f, 30.f, -350.f),
  vec3(60.f, 20.f, -340.f),
  vec3(40.f, 40.f, -400.f)};

vec3 shadNormal = vec3(1.f, 1.f, 1.f);
ShadObj shadower;

/* simple way to extend a point to build shadow volume */
void
mextend(vec3 *newp, vec3 light, vec3 vertex, GLfloat t)
{
  vec3 delta;

  delta.x = vertex.x - light.x;
  delta.y = vertex.y - light.y;
  delta.z = vertex.z - light.z;

  newp->x = light.x + delta.x * t;
  newp->y = light.y + delta.y * t;
  newp->z = light.z + delta.z * t;
}

/* Create a shadow volume in a display list */
gluggModel makeShadowVolume(ShadObj * shadower, vec3 light,
  GLfloat t)
{
  int i;
  vec3 newv;

  gluggBegin(GLUGG_QUADS);
  gluggColor(.2f, .2f, .2f);
  for (i = 0; i < shadower->n; i++)
  {
    gluggVertexv(shadower->verticies[i]);
    mextend(&newv, light, shadower->verticies[i], t);
    gluggVertexv(newv);
    mextend(&newv, light, shadower->verticies[((i + 1) % shadower->n)], t);
    gluggVertexv(newv);
    gluggVertexv(shadower->verticies[((i + 1) % shadower->n)]);
  }
  return gluggBuildModel(0);
}

enum
{
  NONE, NOLIGHT, VOLUME, SHADOW
};

int rendermode = NONE;

void
menu(int mode)
{
  rendermode = mode;
  glutPostRedisplay();
}

GLfloat leftwallshadow[4][4];
GLfloat floorshadow[4][4];

vec3 lightpos = vec3(50.f, 50.f, -340.f);

gluggModel makeFloor()
{
   gluggBegin(GLUGG_QUADS);
   gluggColor(1,1,1);
   gluggNormal(0.f, 1.f, 0.f);
   gluggTexCoord(0, 0);
   gluggVertex(-100.f, -100.f, -320.f);
   gluggTexCoord(1, 0);
   gluggVertex(100.f, -100.f, -320.f);
   gluggTexCoord(1, 1);
   gluggVertex(100.f, -100.f, -520.f);
   gluggTexCoord(0, 1);
   gluggVertex(-100.f, -100.f, -520.f);
   return gluggBuildModel(0);
}

gluggModel makeWalls()
{
  gluggBegin(GLUGG_QUADS);
  /* left wall */
  gluggColor(1,1,1);
  gluggNormal(1.f, 0.f, 0.f);
  gluggVertex(-100.f, -100.f, -320.f);
  gluggVertex(-100.f, -100.f, -520.f);
  gluggVertex(-100.f, 100.f, -520.f);
  gluggVertex(-100.f, 100.f, -320.f);

  /* right wall */
  gluggNormal(-1.f, 0.f, 0.f);
  gluggVertex(100.f, -100.f, -320.f);
  gluggVertex(100.f, 100.f, -320.f);
  gluggVertex(100.f, 100.f, -520.f);
  gluggVertex(100.f, -100.f, -520.f);

  /* ceiling */
  gluggNormal(0.f, -1.f, 0.f);
  gluggVertex(-100.f, 100.f, -320.f);
  gluggVertex(-100.f, 100.f, -520.f);
  gluggVertex(100.f, 100.f, -520.f);
  gluggVertex(100.f, 100.f, -320.f);

  /* back wall */
  gluggNormal(0.f, 0.f, 1.f);
  gluggVertex(-100.f, -100.f, -520.f);
  gluggVertex(100.f, -100.f, -520.f);
  gluggVertex(100.f, 100.f, -520.f);
  gluggVertex(-100.f, 100.f, -520.f);
  return gluggBuildModel(0);
}

gluggModel floorShape;
gluggModel walls;
gluggModel sphere;
gluggModel cone;
gluggModel light;
gluggModel shadow;

GLuint program; // Shader
mat4 projectionMatrix;

/* render while jittering the shadows */
void render(ShadObj * obj)
{
  static GLfloat shad_mat[] = {1.f, .1f, .1f, 1.f};
  int i;

  /* material properties for objects in scene */
  static GLfloat wall_mat[] = {1.f, 1.f, 1.f, 1.f};

  /* Note: wall verticies are ordered so they are all front facing this lets
     me do back face culling to speed things up.  */

  /* floor */
  /* make the floor textured */
  glEnable(GL_TEXTURE_2D);

  /* Since we want to turn texturing on for floor only, we have to make floor 
     a separate glBegin()/glEnd() sequence. You can't turn texturing on and
     off between begin and end calls */
       
	glUniform1i(glGetUniformLocation(program, "enable_texture"), 1);
  	gluggDrawModel(floorShape, program);
	glUniform1i(glGetUniformLocation(program, "enable_texture"), 0);

  glDisable(GL_TEXTURE_2D);

  /* walls */

  	gluggDrawModel(walls, program);
  	
  /* Shapes */
  	
  	gluggDrawModel(cone, program);
  	gluggDrawModel(sphere, program);
  	gluggDrawModel(light, program);

  /* draw shadowing object */
// This is temporary so we can move it!
  gluggBegin(GLUGG_TRIANGLE_FAN);
  gluggNormalv(obj->normal);
  gluggColor(.9f, .2f, .2f);
  for (i = 0; i < obj->n; i++)
  {
    gluggVertexv(obj->verticies[i]);
  }
  gluggModel temp = gluggBuildModel(0);
  gluggDrawModel(temp, program);
  gluggDisposeModel(temp);

}

void redraw(void)
{
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
  switch (rendermode)
  {
  case NONE:
    render(&shadower);
    break;
  case NOLIGHT:
	// Tell shader to make no light
	glUniform1i(glGetUniformLocation(program, "enable_light"), 0);
    render(&shadower);
	glUniform1i(glGetUniformLocation(program, "enable_light"), 1);
    break;
  case VOLUME:
    render(&shadower);
    gluggDrawModel(shadow, program);
    break;
  case SHADOW:
    glColorMask(GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE);
    render(&shadower);  /* render scene in depth buffer */

    glEnable(GL_STENCIL_TEST);
    glDepthMask(GL_FALSE);
    glStencilFunc(GL_ALWAYS, 0, 0);

    glStencilOp(GL_KEEP, GL_KEEP, GL_INCR);
    glCullFace(GL_BACK);  /* increment using front face of shadow volume */
    gluggDrawModel(shadow, program);

    glStencilOp(GL_KEEP, GL_KEEP, GL_DECR);
    glCullFace(GL_FRONT);  /* increment using front face of shadow volume */
    						// Shouldn't that be decrement using back?
    gluggDrawModel(shadow, program);

    glDepthMask(GL_TRUE);
    glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);
    glCullFace(GL_BACK);
    glDepthFunc(GL_LEQUAL);
    glStencilOp(GL_KEEP, GL_KEEP, GL_KEEP);

    glStencilFunc(GL_EQUAL, 1, 1);  /* draw shadowed part */
	glUniform1i(glGetUniformLocation(program, "enable_light"), 0);
    render(&shadower);

    glStencilFunc(GL_EQUAL, 0, 1);  /* draw lit part */
	glUniform1i(glGetUniformLocation(program, "enable_light"), 1);
    render(&shadower);

    glDepthFunc(GL_LESS);
    glDisable(GL_STENCIL_TEST);
    break;
  }

  sgDraw();
  glutSwapBuffers();    /* high end machines may need this */
}

/* ARGSUSED1 */
void key(unsigned char key, int x, int y)
{
  if (key == '\033')
    exit(0);
  if (key == GLUT_KEY_UP)
  {
     shadVerts[0].y += 0.5;
     shadVerts[1].y += 0.5;
     shadVerts[2].y += 0.5;
  }
  if (key == GLUT_KEY_DOWN)
  {
     shadVerts[0].y -= 0.5;
     shadVerts[1].y -= 0.5;
     shadVerts[2].y -= 0.5;
  }
  if (key == GLUT_KEY_LEFT)
  {
     shadVerts[0].x -= 0.5;
     shadVerts[1].x -= 0.5;
     shadVerts[2].x -= 0.5;
  }
  if (key == GLUT_KEY_RIGHT)
  {
     shadVerts[0].x += 0.5;
     shadVerts[1].x += 0.5;
     shadVerts[2].x += 0.5;
  }
  gluggDisposeModel(shadow);
  glutPostRedisplay();
  shadow = makeShadowVolume(&shadower, lightpos, 10.f);
}

void mouse(int button, int state, int x, int y)
{
	if (state == 0) sgMouse(state, x, y);
	glutPostRedisplay();
}

void drag(int x, int y)
{
	sgMouseDrag(x, y);
	glutPostRedisplay();
}


const int TEXDIM = 256;
/* Parse arguments, and set up interface between OpenGL and window system */
int
main(int argc, char *argv[])
{
  GLfloat *tex;
  static GLfloat sphere_mat[] =
  {1.f, .5f, 0.f, 1.f};
  static GLfloat cone_mat[] =
  {0.f, .5f, 1.f, 1.f};

  glutInit(&argc, argv);
  glutInitWindowSize(512, 512);
  glutInitDisplayMode(GLUT_RGBA | GLUT_DEPTH | GLUT_STENCIL | GLUT_DOUBLE);
  glutInitContextVersion(3, 2);
  (void) glutCreateWindow("shadow volumes");
  glutDisplayFunc(redraw);
  glutKeyboardFunc(key);
  	glutMouseFunc(mouse);
	glutMotionFunc(drag);

	program = loadShaders("simple.vert", "simple.frag");

    projectionMatrix = frustum(-100., 100., -100., 100., 320., 640.);
	glUniformMatrix4fv(glGetUniformLocation(program, "projectionMatrix"), 1, GL_TRUE, projectionMatrix.m);

    mat4 cam = lookAt(0, 0, 0, 0, 0, -1, 0,1,0);
    glUniformMatrix4fv(glGetUniformLocation(program, "modelviewMatrix"), 1, GL_TRUE, cam.m);

  /* turn on features */
  glEnable(GL_DEPTH_TEST);
  glEnable(GL_CULL_FACE);
	glUniform1i(glGetUniformLocation(program, "enable_light"), 1);

  /* place light 0 in the right place */
	glUniform3fv(glGetUniformLocation(program, "lightpos"), 1, &lightpos.x);

  /* remove back faces to speed things up */
  glCullFace(GL_BACK);

  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);

  // Make floor and walls
  floorShape = makeFloor();
  walls = makeWalls();

  /* make display lists for sphere and cone; for efficiency */
  
  // I found that I must mirror all models to make it right. I have not yet figured out why,
  // since the same models work correctly in other cases.
  
  gluggBegin(0);
  gluggColor(.9f, .4f, .2f);
  gluggTranslate(60.f, -50.f, -360.f); // moved from the sphere() function
  gluggScale(1.f, 1.f, -1.f); // ??? Bug in gluggShapes?
  gluggSphere(20.f, 20, 20);
  sphere = gluggBuildModel(0);

  gluggBegin(0);
  gluggColor(.2f, .4f, .9f);
  gluggTranslate(-40.f, -40.f, -400.f); // Moved from the cone() function
  gluggScale(1.f, 1.f, -1.f); // ??? Bug in gluggShapes?
  gluggRotate(90.f, 0.f, 1.f, 0.f);
  gluggCone(60, 60.0, 20.0);
  cone = gluggBuildModel(0);
  gluggSetMatrix(IdentityMatrix());
  
  gluggBegin(0);
  gluggTranslate(lightpos.x, lightpos.y, lightpos.z);
  gluggColor(.9f, .9f, .3f);
  gluggScale(1.f, 1.f, -1.f); // ??? Bug in gluggShapes?
  gluggSphere(20, 20, 5.0);
  light = gluggBuildModel(0);
  gluggSetMatrix(IdentityMatrix());

  /* load pattern for current 2d texture */
  tex = make_texture2(TEXDIM, TEXDIM);
  glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, TEXDIM, TEXDIM, 0, GL_RGBA, GL_FLOAT, tex);
  free(tex);
  glUniform1i(glGetUniformLocation(program, "tex"), 0);

	sgSetPosition(20, 20);
	sgCreateRadio(-1, -1, "No shadows", &rendermode, 0);
	sgCreateRadio(-1, -1, "No light", &rendermode, 1);
	sgCreateRadio(-1, -1, "Volume", &rendermode, 2);
	sgCreateRadio(-1, -1, "Shadow", &rendermode, 3);
	sgSetTextColor(0,0,0);
	sgSetFrameColor(0,0,0);
	sgSetBackgroundColor(0.5, 0.5, 0.5, 0.5);


  shadower.verticies = shadVerts;
  shadower.normal = shadNormal;
  shadower.n = sizeof(shadVerts) / (3 * sizeof(GLfloat));

  shadow = makeShadowVolume(&shadower, lightpos, 10.f);

  glutMainLoop();
  return 0;             /* ANSI C requires main to return int. */
}
